/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "ace/POSIX_AIO_LINUX_Provider.h"

#if defined(ACE_HAS_AIO_CALLS) && defined (ACE_HAS_LINUX_LIBAIO) && defined(linux)

#include "ace/Log_Msg.h"

#define POLL_BUF_SIZE 100
/**
 ************************************************************************
 *  LINUX AIO Strategy
 *  based on native LINUX AIO API - io_XXXX
 *
 ************************************************************************
 */

ACE_POSIX_AIO_LINUX_Strategy::~ACE_POSIX_AIO_LINUX_Strategy (void)
{
}

ACE_POSIX_AIO_LINUX_Strategy::ACE_POSIX_AIO_LINUX_Strategy (ACE_POSIX_AIO_LINUX_Provider & provider)
: ACE_POSIX_AIO_Wait_Strategy (provider)
, linux_provider_ (provider)
, next_event_     (0)
, max_num_events_ (sizeof(events_)/sizeof(events_[0]))
, num_events_     (0)
{
	ACE_OS::memset(events_, 0, sizeof(events_));
}


int ACE_POSIX_AIO_LINUX_Strategy::on_aio_start (ACE_POSIX_Asynch_Result * result)
{
  result->iocb_.data = result;  // link to avoid cast
  return 0; 
}


int
ACE_POSIX_AIO_LINUX_Strategy::prepare_to_wait(ACE_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_TRACE ("ACE_POSIX_AIO_LINUX_Strategy::prepare_to_wait");
  
  if (this->num_events_ > 0) 
     return WAIT_COMPLETED;  
  
  next_event_ = 0;
                               
  this->num_events_ = started_list.size ();
  
  if (this->num_events_ > max_num_events_)
    this->num_events_ = max_num_events_;

  return WAIT_ENABLED;
}

int
ACE_POSIX_AIO_LINUX_Strategy::wait (ACE_Time_Value * wait_time)
{
  timespec timeout ;
  timespec * ptime = 0;
  
  if (wait_time != 0) 
    {
      timeout = wait_time->operator timespec();
      ptime = &timeout;
    }
  
  num_events_ = io_getevents (linux_provider_.context_,
                              1,
                              num_events_,
                              events_,
                              ptime);
                              
  
  if (num_events_ > 0)
    {
      next_event_ = events_;
      return WAIT_COMPLETED;
    }
  
  if (this->num_events_ == 0)
    return WAIT_TIMEOUT;

  this->num_events_ = 0;
 
  switch (errno)
    {
    case EAGAIN :   // Timeout
      return WAIT_TIMEOUT;
    case EINTR :    // Interrupted call
      return WAIT_INTERRUPTED;

    default:
          ACE_ERROR((LM_ERROR,
            ACE_LIB_TEXT ("%N:%l:(%P | %t)::%p\n"),
            ACE_LIB_TEXT ("ACE_POSIX_AIO_LINUX_Provider::wait")
            ACE_LIB_TEXT (" <io_getevents> failed\n")));
    }
   return WAIT_ERROR;
}

ACE_POSIX_Asynch_Result *
ACE_POSIX_AIO_LINUX_Strategy::get_next_completed_result 
  (ACE_POSIX_Asynch_Waiting_List & started_list)
{
  ACE_UNUSED_ARG (started_list);
  
  if (this->num_events_ == 0 || next_event_ == 0)
    return 0;
  
  iocb *  ptr_iocb = next_event_->obj;
  void *  data   = next_event_->data;
  u_long  nbytes = next_event_->res;
  u_long  err    = next_event_->res2;
  
  ACE_ASSERT (ptr_iocb->data == data);
  
  ACE_POSIX_Asynch_Result * result =
    ACE_static_cast ( ACE_POSIX_Asynch_Result *, ptr_iocb->data );
  
  ACE_ASSERT ( & result->iocb_ == ptr_iocb);
   
  result->set_ready (nbytes, err, 0);
  
  ++next_event_ ;
  --num_events_ ;
  
  return result;
}

/**
 ************************************************************************
 *  LINUX AIO Provider
 *  based on native LINUX AIO API
 *
 ************************************************************************
 */


ACE_POSIX_AIO_LINUX_Provider::~ACE_POSIX_AIO_LINUX_Provider (void)
{
  io_destroy (context_);
}


ACE_POSIX_AIO_LINUX_Provider::ACE_POSIX_AIO_LINUX_Provider (const ACE_POSIX_AIO_Config & config)
: ACE_POSIX_AIO_Provider (config)
, context_(0)
{
  this->max_aio_num_ = 
    ACE_POSIX_AIO_Provider::get_os_max_aio_num (config.max_num_aio());
    
  int rc = io_setup (ACE_AIO_MAX_SIZE, &context_);

  if (rc != 0)
    ACE_ERROR ((LM_ERROR,
      ACE_LIB_TEXT ("%N:%l:(%P | %t)::%p\n"),
      ACE_LIB_TEXT ("ACE_POSIX_AIO_LINUX_Provider::")
      ACE_LIB_TEXT ("ctor: <io_setup> failed")));
}


int
ACE_POSIX_AIO_LINUX_Provider::get_supported_operations_mask () const
{
  return (ACE_Asynch_Result_Impl::READ_STREAM 
         |ACE_Asynch_Result_Impl::WRITE_STREAM 
         |ACE_Asynch_Result_Impl::READ_FILE 
         |ACE_Asynch_Result_Impl::WRITE_FILE);
}  
  
 
int 
ACE_POSIX_AIO_LINUX_Provider::should_start_interrupt_leader () const
{
  // We do not have to inrterrupt aiowait() 
  // when we start new aio
  return 0;
  // Ignore configuration
  //return ACE_POSIX_AIO_Provider::should_start_interrupt_leader ();
}   

int ACE_POSIX_AIO_LINUX_Provider::can_start_in_any_thread () const
{
  // native API should allow start aio from any thread
  return 1;
  // Ignore configuration
  //return ACE_POSIX_AIO_Provider::can_start_in_any_thread ();
}   

size_t
ACE_POSIX_AIO_LINUX_Provider::get_max_aio_num () const
{
  return this->max_aio_num_; // base class
}

ACE_POSIX_AIO_Wait_Strategy * 
ACE_POSIX_AIO_LINUX_Provider::create_wait_strategy (void)
{
  ACE_POSIX_AIO_Wait_Strategy * wait_strategy = 0;

  ACE_NEW_RETURN (wait_strategy,
                  ACE_POSIX_AIO_LINUX_Strategy (*this),
                  0);

  return wait_strategy;
}

void
ACE_POSIX_AIO_LINUX_Provider::destroy_wait_strategy (ACE_POSIX_AIO_Wait_Strategy * w_strategy)
{
  delete  w_strategy;
}

int
ACE_POSIX_AIO_LINUX_Provider::register_handle (ACE_HANDLE handle,
                                             const void *completion_key)
{
  ACE_UNUSED_ARG (handle);
  ACE_UNUSED_ARG (completion_key);

  return 0;
}

int
ACE_POSIX_AIO_LINUX_Provider::unregister_handle (ACE_HANDLE handle)
{
  ACE_UNUSED_ARG (handle);

  return 0;
}

int
ACE_POSIX_AIO_LINUX_Provider::start_read_stream (ACE_POSIX_Asynch_Result *result)
{  
  io_prep_pread (&result->iocb_,
                 result->aio_fildes,
                 (void*) result->aio_buf,
                 result->aio_nbytes,
                 result->aio_offset);

  result->iocb_.data = result;  // link to avoid cast
  
  
  iocb *  ptr = &result->iocb_;
  int rc = io_submit (context_, 1, &ptr);
  
  this->check_start_error (result, rc);

  return 0;
}

int
ACE_POSIX_AIO_LINUX_Provider::start_write_stream (ACE_POSIX_Asynch_Result *result)
{  
  io_prep_pwrite (&result->iocb_,
                 result->aio_fildes,
                 (void*) result->aio_buf,
                 result->aio_nbytes,
                 result->aio_offset);


  result->iocb_.data = result;  // link to avoid cast
  
  
  iocb *  ptr = &result->iocb_;
  int rc = io_submit (context_, 1, &ptr);
  
  this->check_start_error (result, rc);

  return 0;
}

int
ACE_POSIX_AIO_LINUX_Provider::start_read_file (ACE_POSIX_Asynch_Result *result)
{  
  return this->start_read_stream (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::start_write_file (ACE_POSIX_Asynch_Result *result)
{  
  return this->start_write_stream (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::start_read_dgram (ACE_POSIX_Asynch_Result *result)
{
  return ACE_POSIX_AIO_Provider::start_not_supported (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::start_write_dgram (ACE_POSIX_Asynch_Result *result)
{  
  return ACE_POSIX_AIO_Provider::start_not_supported (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::start_accept (ACE_POSIX_Asynch_Result *result)
{
  return ACE_POSIX_AIO_Provider::start_not_supported (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::start_connect (ACE_POSIX_Asynch_Result *result)
{  
  return ACE_POSIX_AIO_Provider::start_not_supported (result);
}

int
ACE_POSIX_AIO_LINUX_Provider::cancel_aio (ACE_POSIX_Asynch_Result * result)
{
  /// Cancel pending AIO request

  ACE_TRACE ("ACE_POSIX_AIO_LINUX_Provider::cancel_aio");
  
  io_event event;

  int rc = ::io_cancel (context_, 
                        &result->iocb_,
                        &event);
  if (rc == 0)       // OK, canceled and dequeued
    {
      result->set_ready (event.res, event.res2, 0);
    
      return CR_CANCELED;
    }
  
  switch ( errno )  
    {
    case EAGAIN:
      return CR_NOT_CANCELED;
      
    default:
      break;
    }

  return CR_CANCEL_ERROR;
}


#endif /* ACE_HAS_AIO_CALLS && ACE_HAS_LINUX_LIBAIO && linux */
