/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id: $

#include "ace/POSIX_AIO_Processor.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/POSIX_AIO_Provider.h"
#include "ace/POSIX_AIO_Dispatcher.h"

#include "ace/ACE.h"
#include "ace/Log_Msg.h"

typedef ACE_Reverse_Lock<ACE_SYNCH_MUTEX> REVERSE_MUTEX;

#define MAX_NUM_NOTIFY 1

/*
 *******************************************************************
 * ACE_POSIX_AIO_Processor
 *******************************************************************
 */
ACE_POSIX_AIO_Processor::ACE_POSIX_AIO_Processor
  (ACE_POSIX_AIO_Dispatcher & completion_dispatcher,
   const ACE_POSIX_AIO_Config & config)
  : config_                (config),
    completion_dispatcher_ (completion_dispatcher),
    mutex_                 (completion_dispatcher.mutex()),
    cancel_condition_      (mutex_),
    aio_provider_          (0),
    wait_strategy_         (0),
    interrupt_strategy_    (0),
    interrupt_state_       (0),
    leader_id_             (ACE_OS::NULL_thread)
{
  ;
}
 

ACE_POSIX_AIO_Processor::~ACE_POSIX_AIO_Processor (void)
{
  ACE_ASSERT (this->aio_provider_ == 0);
}

/*
 *******************************************************************
 * dump
 *******************************************************************
 */
int 
ACE_POSIX_AIO_Processor::dump ()
{
  ACE_POSIX_Asynch_Result * result = 0;

  ACE_DEBUG((LM_DEBUG,
             ACE_LIB_TEXT("Deferred list size=%d\n"),
             this->deferred_list_.size()));

  ACE_POSIX_Asynch_Result_List_Iterator itr (this->deferred_list_);
  

  while ((result=itr.get_current_and_advance ()) != 0)
    result->dump();

  ACE_DEBUG((LM_DEBUG,
             ACE_LIB_TEXT("Started list size=%d\n"),
             this->started_list_.size()));

  itr.reset (this->started_list_);

  while ((result=itr.get_current_and_advance ()) != 0)
    result->dump();

  return 0;
}

int
ACE_POSIX_AIO_Processor::init_provider ()
{
  if (this->wait_strategy_->init() != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT ("(%t) %p\n"),
      ACE_LIB_TEXT ("ACE_POSIX_AIO_Wait_Strategy::init() failed")),
      -1);

  if (this->interrupt_strategy_->init() != 0)
    ACE_ERROR_RETURN ((LM_ERROR,
      ACE_LIB_TEXT ("(%t) %p\n"),
      ACE_LIB_TEXT ("ACE_POSIX_AIO_Interrupt_Strategy::init() failed")),
      -1);

  return 0;
}

int
ACE_POSIX_AIO_Processor::fini_provider ()
{
  if (this->interrupt_strategy_ != 0)
    this->interrupt_strategy_->interrupt (ACE_Thread::self());

  if (this->aio_provider_ != 0)
    {
      this->cancel_all_deferred_aio ();

      int num_try = 10;

      while (this->started_list_.size () > 0 && num_try > 0)
        {
          this->cancel_all_started_aio ();

          this->poll_events ();
        
          num_try--;
        }

      if (this->started_list_.size () > 0)
        {
          this->dump();
        
          // we can not delete uncompleted results
          // we have to lost them

          ACE_POSIX_Asynch_Result * result = 0;

          while ((result = this->started_list_.get_first()) != 0)
            this->started_list_.remove_result (result);
        }

    }

  if (this->interrupt_strategy_ != 0)
    this->interrupt_strategy_->fini ();

  if (this->wait_strategy_ != 0)
    this->wait_strategy_->fini ();

  return 0;
}

int
ACE_POSIX_AIO_Processor::create_provider ()
{
  if (this->aio_provider_ != 0)
    return -1;

  this->aio_provider_ = this->config_.create_provider ();

  if (this->aio_provider_ == 0)
       return -1;

  if (this->aio_provider_->create_strategies () != 0) 
    {
      this->destroy_provider ();
      return -1;
    }
    
  this->wait_strategy_ = this->aio_provider_->wait_strategy ();
  this->interrupt_strategy_ = this->aio_provider_->interrupt_strategy ();

  return 0;
}

int
ACE_POSIX_AIO_Processor::destroy_provider ()
{
  if (this->aio_provider_ != 0)
    {
      this->aio_provider_->destroy_strategies ();

      delete this->aio_provider_;
    }

  this->aio_provider_ =0;
  this->wait_strategy_ = 0;
  this->interrupt_strategy_ = 0;
  
  return 0;
}

int
ACE_POSIX_AIO_Processor::is_supported (int op) const
{
  int mask = this->aio_provider_->get_supported_operations_mask ();

  return (ACE_BIT_ENABLED (mask, op)) ? 1 : 0;
}

/**
 ****************************************************************
 * virtual methods 
 ****************************************************************
 */

int 
ACE_POSIX_AIO_Processor::start (void)
{
  return this->init_provider ();
}

int 
ACE_POSIX_AIO_Processor::stop (void)
{
  return this->fini_provider ();
}

/**
 ****************************************************************
 * virtual  Leadership methods
 ****************************************************************
 */

int 
ACE_POSIX_AIO_Processor::is_dedicated_leadership () const
{
  return 0;
}  
int 
ACE_POSIX_AIO_Processor::is_ready_for_leadership () const
{
  return (this->num_total_aios() != 0);
}

int
ACE_POSIX_AIO_Processor::activate_leader (int reason)
{
  if (this->has_leader ())
    return this->interrupt_leader(reason);

  this->completion_dispatcher_.notify (ACE_POSIX_AIO_Dispatcher::NOTIFY_FOLLOWER);

  return 0;
}

/**
 ****************************************************************
 * Test,Acquire and Release Leadership methods
 ****************************************************************
 */

size_t
ACE_POSIX_AIO_Processor::num_total_aios () const
{
  return (this->started_list_.size () + 
          this->deferred_list_.size ());
}

int 
ACE_POSIX_AIO_Processor::has_leader() const
{
  return !(ACE_OS::thr_equal (this->leader_id_, ACE_OS::NULL_thread));
}

int 
ACE_POSIX_AIO_Processor::is_in_leader_context() const 
{
  return ACE_OS::thr_equal (ACE_Thread::self(), this->leader_id_);
}

int 
ACE_POSIX_AIO_Processor::has_shared_leader() const
{
  return (!this->is_dedicated_leadership() && has_leader());
}
  
int 
ACE_POSIX_AIO_Processor::has_dedicated_leader() const
{
  return (this->is_dedicated_leadership() && has_leader());
}

int 
ACE_POSIX_AIO_Processor::is_in_shared_leader_context() const
{
  return (!this->is_dedicated_leadership() && is_in_leader_context());
}

int 
ACE_POSIX_AIO_Processor::is_in_dedicated_leader_context () const
{
  return (this->is_dedicated_leadership() && is_in_leader_context());
}

int 
ACE_POSIX_AIO_Processor::acquire_leadership ()
{
  if (this->has_leader())
    return 0;

  // ask is we are ready, i.e. able to do useful job
  if (!is_ready_for_leadership())
    return 0;

  this->leader_id_ = ACE_Thread::self ();

  return 1;
}

int 
ACE_POSIX_AIO_Processor::release_leadership ()
{
  if (!this->is_in_leader_context())
    return -1;
 
  this->leader_id_ = ACE_OS::NULL_thread;

  return 0;
}  

int 
ACE_POSIX_AIO_Processor::interrupt_leader (int reason)
{
  //this method should not be called in leader context
  //ACE_ASSERT (!this->is_in_leader_context());

  if (!this->has_leader ())
    return 0;

  int flg = (this->interrupt_state_ != 0);

  this->interrupt_state_ |= reason;

  if (flg)
    return 0;
  

  return this->interrupt_strategy_->interrupt (this->leader_id_);
}


/*
 *******************************************************************
 * 
 *
 *******************************************************************
 */

int 
ACE_POSIX_AIO_Processor::post_result (ACE_POSIX_Asynch_Result * result,
                                      int notify_flg)
{
  // Optimization:
  // Do not waste time to notify completion dispatcher,
  // if we are in the leader context.
  // Later, when we finish handle events,
  // we will notify the completion dispatcher
   
  if (result->is_internal_result ())
   {

      // Call completion hook
      result->complete (result->bytes_transferred (),
                        result->success (),
                        result->completion_key (),
                        result->error ());

      return 0;  // do nothing;
   }

  
  if (this->is_dedicated_leadership ())
    notify_flg = ACE_POSIX_AIO_Dispatcher::NOTIFY_FOLLOWER;

  this->completion_dispatcher_.post_result (result, notify_flg);
  return 1;
}


int 
ACE_POSIX_AIO_Processor::handle_aio_events (ACE_Time_Value * wait_time)
{
  //ACE_ASSERT (this->is_in_leader_context());

  int rc = 0;

  for (;;)
    {
      if ((this->interrupt_state_ & IR_CANCEL_AIO) != 0)
        {
          this->interrupt_state_ &= ~IR_CANCEL_AIO;
          this->interrupt_state_ &= ~IR_BREAK;

          this->cancel_condition_.broadcast();
          break;
        }

      if ((this->interrupt_state_ & IR_BREAK) != 0)
        {
          this->interrupt_state_ &= ~IR_BREAK;

          break;
        }


      //activate interrupt strategy
      this->interrupt_strategy_->on_leader_activated (*this);

      rc = this->handle_aio_events_i (wait_time);

      //deactivate interrupt strategy
      this->interrupt_strategy_->on_leader_deactivated (*this);

      if (rc != 0)  // we had completed results or errors
        {
          // notify completion dispatcher if leadership is dedicated
      
          //    if (this->is_dedicated_leadership())
          //      this->completion_dispatcher_.notify 
          //        (ACE_POSIX_AIO_Dispatcher::NOTIFY_FOLLOWER);
          
          this->interrupt_state_ |= IR_BREAK;
        }

      // else rc == 0, no events , 
      // may be we were interrupted to update wait state

      if (wait_time != 0 && *wait_time == ACE_Time_Value::zero)
        {
          this->interrupt_state_ |= IR_BREAK;
        }
    }

  //this->interrupt_state_ &= ~IR_BREAK;

  return rc;
}


int 
ACE_POSIX_AIO_Processor::handle_aio_events_i (ACE_Time_Value * wait_time)
{
  // Auto decrement wait time
  ACE_Countdown_Time countdown (wait_time);

  //clear interrupt state flags
  this->interrupt_state_ &= ~(IR_NEW_DEFERRED_AIO | IR_NEW_STARTED_AIO);


  int rc = this->start_deferred_aio ();
  if (rc < 0)    // errors
    return rc;

  if (rc > 0)    // some completions
    {
      int rc_poll = this->poll_events ();

      if (rc_poll >= 0)
        rc_poll += rc;

      return rc_poll;
    }


  rc = this->wait_strategy_->prepare_to_wait (this->started_list_);

  if (rc == ACE_POSIX_AIO_Wait_Strategy::WAIT_ENABLED)
  {
    // All methods of ACE_POSIX_AIO_Processor 
    // are called with locked Proactor Mutex.

    // This is only one place where
    // we  should release this mutex for time waiting in:
    // aio_suspend()/aiowait()/select()
    // For this purpose let have ACE_Reverse_Lock<>

    REVERSE_MUTEX anti_mutex (this->mutex_);
    ACE_MT (ACE_GUARD_RETURN (REVERSE_MUTEX, ace_mon, anti_mutex, -1));

    rc = this->wait_strategy_->wait (wait_time);

  } // Release anti-mutex, i.e obtain mutex

  switch (rc)
    {
    case ACE_POSIX_AIO_Wait_Strategy::WAIT_COMPLETED:
      break;

    case ACE_POSIX_AIO_Wait_Strategy::WAIT_TIMEOUT:
    case ACE_POSIX_AIO_Wait_Strategy::WAIT_INTERRUPTED:
      break;

    case ACE_POSIX_AIO_Wait_Strategy::WAIT_ERROR:
    default:
      return -1;
    }


  ACE_POSIX_Asynch_Result * result=0;

  for (rc=0;;)
    {
      result = this->wait_strategy_->
                 get_next_completed_result (this->started_list_);
      
      if (!result)
        break;

      //remove it from started_list_
      this->started_list_.remove_result (result);

      // pass result to completion_dispatcher_
      if (this->post_result (result,
                             ACE_POSIX_AIO_Dispatcher::NOTIFY_NONE) > 0)
        {
          rc++;
        }

    }

  return rc;
}

int 
ACE_POSIX_AIO_Processor::poll_events (void)
{
  ACE_Time_Value tv = ACE_Time_Value::zero;

  int rc = this->wait_strategy_->prepare_to_wait (this->started_list_);

  if (rc == ACE_POSIX_AIO_Wait_Strategy::WAIT_ENABLED)
    rc = this->wait_strategy_->wait (&tv);

  switch (rc)
    {
    case ACE_POSIX_AIO_Wait_Strategy::WAIT_COMPLETED:
      break;

    case ACE_POSIX_AIO_Wait_Strategy::WAIT_TIMEOUT:
    case ACE_POSIX_AIO_Wait_Strategy::WAIT_INTERRUPTED:
      break;

    case ACE_POSIX_AIO_Wait_Strategy::WAIT_ERROR:
    default:
      return -1;
    }

  ACE_POSIX_Asynch_Result * result=0;

  for (rc=0;;)
    {
      result = this->wait_strategy_->
                 get_next_completed_result (this->started_list_);
      
      if (!result)
        break;

      //remove it from started_list_
      this->started_list_.remove_result (result);

      // pass result to completion_dispatcher_
      if (this->post_result (result,
                             ACE_POSIX_AIO_Dispatcher::NOTIFY_NONE) > 0)
        {
          rc++;
        }
    }

  return rc;
}

/*
 *******************************************************************
 * group of start aio methods
 *
 *******************************************************************
 */

void
ACE_POSIX_AIO_Processor::report_start_error_i (ACE_POSIX_Asynch_Result *result)
{
    ACE_ERROR ((LM_ERROR,
      ACE_LIB_TEXT ("%N:%l:(%P|%t) ACE_POSIX_AIO_Processor::")
      ACE_LIB_TEXT ("start_aio: %p\n"),
      ACE_LIB_TEXT ("failed\n")));

    result->dump();
}
    

int
ACE_POSIX_AIO_Processor::start_aio (ACE_POSIX_Asynch_Result * result)
{
  ACE_TRACE ("ACE_POSIX_AIO_Processor::start_aio");

  // this method is never called in leader context


  // check that we are able to start
  if (this->started_list_.size () >= this->aio_provider_->get_max_aio_num ())
    {
      result->set_deferred ();
      this->deferred_list_.insert_tail (result);
      return 0;
    }

  // check if we can start aio from any thread
  if (!this->aio_provider_->can_start_in_any_thread ())
    {
      // Start enable on only in leader context,
      // so add result to the deferred queue 
      // and wake up leader to start new deferred aios.
      // If number total aios is 1,
      // then we should activate leader.
      // For shared leadership wakeup completion_dispatcher
      // to activate leader
      // For dedicated - wake up it from sleep.
    
      result->set_deferred ();
      this->deferred_list_.insert_tail (result);

      if (this->num_total_aios () == 1) 
        this->activate_leader (IR_NEW_DEFERRED_AIO);
      else
        this->interrupt_leader (IR_NEW_DEFERRED_AIO); 
      
      return 0;
    }

  // start in aio current, i.e. non leader thread
  // on return ret_val contains:
  //  1 - we have to interrupt leader,
  //      wait state should be updated
  //  0 - we do not have to interrupt leader

  int ret_val = this->aio_provider_->start_aio (result);

  switch (result->state ())
    {
    case ACE_Asynch_Result_Impl::STARTED:       // started OK 
      this->started_list_.add_result (result);
      break;

    case ACE_Asynch_Result_Impl::DEFERRED:      // try again later
      this->deferred_list_.insert_tail (result); // add to tail
      break;

    case ACE_Asynch_Result_Impl::RESULT_READY:   // already finished, 
      this->post_result (result,
                         ACE_POSIX_AIO_Dispatcher::NOTIFY_STRONG);
      return 0;

    case ACE_Asynch_Result_Impl::START_ERROR:   // Invalid Parameters 
      this->report_start_error_i (result);
      return -1;

    default :       
      this->report_start_error_i (result);
      ACE_ASSERT (0);
      return -1;   
    }
  
  // here we are if start was OK
  // if necessary, wake up the leader to renew wait state   

  // If number total aios is 1,
  // then we should activate leader.
  // For shared leadership wakeup completion_dispatcher
  // to activate leader
  // For dedicated - wake up the leader from sleep.

  if (this->num_total_aios () == 1)
    {
      this->activate_leader (IR_NEW_STARTED_AIO);
      return 0;
    }

  if (ret_val == 0)
    return 0;                                  

  if (!this->has_leader ())
    return 0;
  
  if (!this->completion_dispatcher_.is_handle_in_dispatch (result->aio_fildes,
                                                           this))
    this->interrupt_leader (IR_NEW_STARTED_AIO);

  return 0;
}

int
ACE_POSIX_AIO_Processor::start_deferred_aio ()
{
  // this method is always called in leader context

  ACE_TRACE ("ACE_POSIX_AIO_Processor::start_deferred_aio");

  int num_completed = 0;
  size_t limit = this->aio_provider_->get_max_aio_num ();

  ACE_POSIX_Asynch_Result_List_Iterator  itr(this->deferred_list_);

  itr.first();
  
  while (this->started_list_.size () < limit )
    { 
      ACE_POSIX_Asynch_Result * result = itr.get_current_and_advance ();

      if (!result)
        break;

      this->aio_provider_->start_aio (result);

      switch (result->state())
        {
        case ACE_Asynch_Result_Impl::STARTED:        // started OK 
          this->deferred_list_.remove (result);
          this->started_list_.add_result (result);
          break;

        case ACE_Asynch_Result_Impl::DEFERRED:       // try again later
          break;

        case ACE_Asynch_Result_Impl::RESULT_READY:   // finished,
          this->deferred_list_.remove (result);
          this->post_result (result,
                             ACE_POSIX_AIO_Dispatcher::NOTIFY_NONE);
          ++num_completed;
          break;

        case ACE_Asynch_Result_Impl::START_ERROR:    // Invalid Parameters
          this->deferred_list_.remove (result);
          this->report_start_error_i(result);
          this->post_result (result,
                             ACE_POSIX_AIO_Dispatcher::NOTIFY_NONE);
          ++num_completed;
          break;

        default :                                   
          this->deferred_list_.remove (result);
          this->report_start_error_i(result);
          this->post_result (result,
                             ACE_POSIX_AIO_Dispatcher::NOTIFY_NONE);
          ++num_completed;
          ACE_ASSERT (0);
          return -1;
        }

    }
        
  return num_completed;
}

int
ACE_POSIX_AIO_Processor::start_special_aio (ACE_POSIX_Asynch_Result * result)
{
  result->set_deferred ();
  this->deferred_list_.insert_head (result); // add to head

  return 0;
}
  
/**
 *******************************************************************
 *  Group of cancel methods
 *
 *******************************************************************
 */
 
int
ACE_POSIX_AIO_Processor::cancel_all_aio (void)
{
  // For all deferred AIO requests with handle "handle"
  // it removes its from the lists and notifies user
  //
  // For all running AIO requests with handle "handle"
  // it calls ::aio_cancel. According to the POSIX standards
  // we will receive ECANCELED  for all ::aio_canceled AIO requests
  // later on return from ::aio_suspend
  //

  ACE_TRACE ("ACE_POSIX_AIO_Processor::cancel_all_aio");

  int rc1 = this->cancel_all_deferred_aio ();
  int rc2 = this->cancel_all_started_aio ();

  if (rc2 == ACE_POSIX_AIO_Provider::CR_ALL_DONE)
    return rc1;

  return rc2;
} 
int
ACE_POSIX_AIO_Processor::cancel_aio (ACE_HANDLE handle)
{
  // For all deferred AIO requests with handle "handle"
  // it removes its from the lists and notifies user
  //
  // For all running AIO requests with handle "handle"
  // it calls ::aio_cancel. According to the POSIX standards
  // we will receive ECANCELED  for all ::aio_canceled AIO requests
  // later on return from ::aio_suspend
  //

  ACE_TRACE ("ACE_POSIX_AIO_Processor::cancel_aio handle");

  int rc1 = this->cancel_deferred_aio (handle);
  int rc2 = this->cancel_started_aio (handle);

  if (rc2 == ACE_POSIX_AIO_Provider::CR_ALL_DONE)
    return rc1;

  return rc2;
}

int
ACE_POSIX_AIO_Processor::cancel_aio (ACE_Handler * handler)
{
  // For all deferred AIO requests with handle "handle"
  // it removes its from the lists and notifies user
  //
  // For all running AIO requests with handle "handle"
  // it calls ::aio_cancel. According to the POSIX standards
  // we will receive ECANCELED  for all ::aio_canceled AIO requests
  // later on return from ::aio_suspend
  //

  ACE_TRACE ("ACE_POSIX_AIO_Processor::cancel_aio handler");

  int rc1 = this->cancel_deferred_aio (handler);
  int rc2 = this->cancel_started_aio (handler);

  if (rc2 == ACE_POSIX_AIO_Provider::CR_ALL_DONE)
    return rc1;

  return rc2;
}


int
ACE_POSIX_AIO_Processor::cancel_all_deferred_aio (void)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->deferred_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;

  while ((result=itr.get_current_and_advance ()) != 0)
    {
      this->cancel_deferred_aio (result);
      ++num_canceled;
    }

  if (num_canceled > 0) 
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_ALL_DONE;
}

int
ACE_POSIX_AIO_Processor::cancel_deferred_aio (ACE_Handler * handler)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->deferred_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;
 
  while ((result=itr.get_current_and_advance ()) != 0)
    {
      if (handler != & result->handler())  // not our aio 
        continue;

      this->cancel_deferred_aio (result);
      ++num_canceled;
    }

  if (num_canceled > 0) 
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_ALL_DONE;
}

int
ACE_POSIX_AIO_Processor::cancel_deferred_aio (ACE_HANDLE handle)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->deferred_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;
 
  while ((result=itr.get_current_and_advance ()) != 0)
    {
      if (result->aio_fildes != handle)  // not our aio 
        continue;

      this->cancel_deferred_aio (result);
      ++num_canceled;
    }

  if (num_canceled > 0) 
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_ALL_DONE;
}


int
ACE_POSIX_AIO_Processor::cancel_deferred_aio (ACE_POSIX_Asynch_Result * result)
{
  ACE_ASSERT (result->state ()== ACE_Asynch_Result_Impl::DEFERRED);

  // now mark result as ready
  result->set_ready (0, ECANCELED, 0);

  //remove it from deferred_list_
  this->deferred_list_.remove (result);

  // pass result to the completion_dispatcher_
  this->post_result (result,
                     ACE_POSIX_AIO_Dispatcher::NOTIFY_STRONG);


  return ACE_POSIX_AIO_Provider::CR_CANCELED;
}

int
ACE_POSIX_AIO_Processor::cancel_all_started_aio (void)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->started_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;
  int num_total = 0;
 
  while ((result=itr.get_current_and_advance ()) != 0)
    {
      ++num_total;

      ACE::set_flags (result->aio_fildes, ACE_NONBLOCK);

      switch (this->cancel_started_aio (result))
        {
        case ACE_POSIX_AIO_Provider::CR_CANCELED:
        case ACE_POSIX_AIO_Provider::CR_ALL_DONE:
          ++num_canceled;
        }

    }

  if (num_total == 0)
    return ACE_POSIX_AIO_Provider::CR_ALL_DONE;

  this->interrupt_leader_and_wait();

  if (num_canceled == num_total)
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_NOT_CANCELED;
}

int
ACE_POSIX_AIO_Processor::cancel_started_aio (ACE_Handler * handler)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->started_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;
  int num_total = 0;
 
  while ((result=itr.get_current_and_advance ()) != 0)
    {
      if (handler != & result->handler())  // not our aio 
        continue;

      ++num_total;

      switch (this->cancel_started_aio (result))
        {
        case ACE_POSIX_AIO_Provider::CR_CANCELED:
        case ACE_POSIX_AIO_Provider::CR_ALL_DONE:
          ++num_canceled;
        }
    }

  if (num_total == 0)
    return ACE_POSIX_AIO_Provider::CR_ALL_DONE;

  this->interrupt_leader_and_wait();

  if (num_canceled == num_total)
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_NOT_CANCELED;
}

int
ACE_POSIX_AIO_Processor::cancel_started_aio (ACE_HANDLE handle)
{
  ACE_POSIX_Asynch_Result_List_Iterator itr (this->started_list_);
  
  ACE_POSIX_Asynch_Result * result = 0;

  int num_canceled = 0;
  int num_total = 0;
 
  while ((result=itr.get_current_and_advance ()) != 0)
    {
      if (result->aio_fildes != handle)  // not our aio 
        continue;

      ++num_total;

      switch (this->cancel_started_aio (result))
        {
        case ACE_POSIX_AIO_Provider::CR_CANCELED:
        case ACE_POSIX_AIO_Provider::CR_ALL_DONE:
          ++num_canceled;
        }
    }

  if (num_total == 0)
    return ACE_POSIX_AIO_Provider::CR_ALL_DONE;

  this->interrupt_leader_and_wait();

  if (num_canceled == num_total)
    return ACE_POSIX_AIO_Provider::CR_CANCELED;

  return ACE_POSIX_AIO_Provider::CR_NOT_CANCELED;
}

int
ACE_POSIX_AIO_Processor::cancel_started_aio (ACE_POSIX_Asynch_Result * result)
{
  // ACE_ASSERT (result->state ()== ACE_Asynch_Result_Impl::STARTED);

  int rc = this->aio_provider_->cancel_aio (result);

  switch (result->state())
    {
    case ACE_Asynch_Result_Impl::START_ERROR:
    case ACE_Asynch_Result_Impl::RESULT_READY:
      this->started_list_.remove_result (result);
      this->post_result (result,
                         ACE_POSIX_AIO_Dispatcher::NOTIFY_STRONG);
      break;

    case ACE_Asynch_Result_Impl::STARTED:
      break;

    default:
      ACE_ASSERT (0);
      break;
    }

  return rc;
}

int
ACE_POSIX_AIO_Processor::interrupt_leader_and_wait (void)
{
  this->interrupt_leader (IR_CANCEL_AIO);

  while ((this->interrupt_state_ & IR_CANCEL_AIO) != 0)
    {
      this->cancel_condition_.wait();
    }

  return 0;
}
#endif /* ACE_HAS_AIO_CALLS */
