/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */
// $Id:  $

#include "ace/POSIX_AIO_Provider.h"

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)

#include "ace/Log_Msg.h"
#include "ace/ACE.h"

# if defined (ACE_HAS_SYSINFO)
#   include  <sys/systeminfo.h>
# endif /* ACE_HAS_SYS_INFO */


ACE_POSIX_AIO_Provider::ACE_POSIX_AIO_Provider (const ACE_POSIX_AIO_Config & config)
  : config_      (config),
    max_aio_num_ (config.max_num_aio ()),
    wait_strategy_ (0),
    interrupt_strategy_ (0)
{
  ;
}

ACE_POSIX_AIO_Provider::~ACE_POSIX_AIO_Provider (void)                                            
{
  ACE_ASSERT(this->wait_strategy_ == 0);
  ACE_ASSERT(this->interrupt_strategy_ == 0);
}

const ACE_POSIX_AIO_Config & 
ACE_POSIX_AIO_Provider::config () const
{
  return this->config_;
}

ACE_POSIX_AIO_Wait_Strategy *
ACE_POSIX_AIO_Provider::wait_strategy () const
{
  return this->wait_strategy_;
}

ACE_POSIX_AIO_Interrupt_Strategy * 
ACE_POSIX_AIO_Provider::interrupt_strategy () const
{
  return this->interrupt_strategy_;
}

int 
ACE_POSIX_AIO_Provider::create_strategies (void)
{
  if (this->wait_strategy_ == 0)
    this->wait_strategy_ = this->create_wait_strategy ();

  if (this->wait_strategy_ == 0)
    return -1;

  if (this->interrupt_strategy_ == 0)
    this->interrupt_strategy_ = 
      this->wait_strategy_->create_interrupt_strategy ();

  if (this->interrupt_strategy_ == 0)
    return -1;

  return 0;
}

int
ACE_POSIX_AIO_Provider::destroy_strategies (void)
{
  if (this->wait_strategy_ != 0)
    {
      if (this->interrupt_strategy_ != 0)
      {
        this->wait_strategy_->destroy_interrupt_strategy
                                (this->interrupt_strategy_);

        this->interrupt_strategy_ = 0;
      }

      this->destroy_wait_strategy (this->wait_strategy_);

      this->wait_strategy_ =0;
    }

  return 0;
}

int ACE_POSIX_AIO_Provider::should_start_interrupt_leader () const
{
  return ACE_BIT_ENABLED (this->config_.provider_flags(),
                          ACE_POSIX_AIO_Config::PVF_AIO_START_INTERRUPT_WAIT);
}

int ACE_POSIX_AIO_Provider::can_start_in_any_thread () const
{
  return ACE_BIT_ENABLED (this->config_.provider_flags(),
                          ACE_POSIX_AIO_Config::PVF_AIO_START_ANY_THREAD);
}

size_t ACE_POSIX_AIO_Provider::get_max_aio_num () const
{
  return this->max_aio_num_;
}


size_t 
ACE_POSIX_AIO_Provider::get_os_max_aio_num (size_t max_op )
{
  // check for user-defined value
  // ACE_AIO_MAX_SIZE if defined in POSIX_Proactor.h

  if (max_op == 0 || max_op > ACE_AIO_MAX_SIZE)
     max_op = ACE_AIO_MAX_SIZE;

  long max_os_aio_num = ACE_OS::sysconf (_SC_AIO_MAX);

  // Define max limit AIO's for concrete OS
  // -1 means that there is no limit, but it is not true
  // (example, SunOS 5.6)

  if (max_os_aio_num > 0 &&
      max_op > (unsigned long) max_os_aio_num)
     max_op = max_os_aio_num;

#if defined (HPUX)
  // Although HPUX 11.00 allows to start 2048 AIO's
  // for all process in system
  // it has a limit 256 max elements for aio_suspend ()
  // It is a pity, but ...

  long max_os_listio_num = ACE_OS::sysconf (_SC_AIO_LISTIO_MAX);
  if (max_os_listio_num > 0
      && max_op > (unsigned long) max_os_listio_num)
    max_op = max_os_listio_num;
#endif /* HPUX */

  // check for max number files to open

  int max_num_files = ACE::max_handles ();

  if (max_num_files > 0
      && max_op > (unsigned long) max_num_files)
    {
      ACE::set_handle_limit (max_op);

      max_num_files = ACE::max_handles ();
    }

  if (max_num_files > 0
      && max_op > (unsigned long) max_num_files)
    max_op = (unsigned long) max_num_files;

  ACE_DEBUG((LM_DEBUG,
    ACE_LIB_TEXT("(%P|%t) ACE_POSIX_AIO_Provider::")
    ACE_LIB_TEXT("Max Number of AIOs=%d\n"),
    max_op));

  return max_op;
}

void
ACE_POSIX_AIO_Provider::check_start_error (ACE_POSIX_Asynch_Result *result,
                                           int rc)
{  
  if (rc >= 0)  //  success on start
    return;

  switch (ACE_OS::last_error())
    {
    case EAGAIN:
    case ENOMEM:  //Ok, it will be deferred AIO
      result->set_deferred ();
      break;
      
    default:
      result->set_start_error (errno);
      break;
    }
  return;
}

int
ACE_POSIX_AIO_Provider::start_not_supported (ACE_POSIX_Asynch_Result *result)
{
  ACE_OS::last_error (EINVAL);  
  ACE_POSIX_AIO_Provider::check_start_error (result, -1);

  return 0;  // don't need interrupt leader
}

int
ACE_POSIX_AIO_Provider::start_aio (ACE_POSIX_Asynch_Result *result)
{  
  // this method is internal and private
  ACE_TRACE ("ACE_POSIX_AIO_Provider::start_aio");


  result->set_started ();

  int ret_val = this->wait_strategy_->on_aio_start (result);
  
  //check if state was changed, may be deferred op
  if (result->state() != ACE_Asynch_Result_Impl::STARTED)
    return ret_val;

  //if (ret_val < 0)
  //  return ACE_POSIX_AIO_Provider::start_not_supported (result);


  switch (result->op_code ())
    {
    case ACE_Asynch_Result_Impl::READ_STREAM:
      ret_val = this->start_read_stream (result);
      break;

    case ACE_Asynch_Result_Impl::WRITE_STREAM:
      ret_val = this->start_write_stream (result);
      break;

    case ACE_Asynch_Result_Impl::READ_FILE:
      ret_val = this->start_read_file (result);
      break;

    case ACE_Asynch_Result_Impl::WRITE_FILE:
      ret_val = this->start_write_file (result);
      break;

    case ACE_Asynch_Result_Impl::ACCEPT:
      ret_val = this->start_accept (result);
      break;

    case ACE_Asynch_Result_Impl::CONNECT:
      ret_val = this->start_connect (result);
      break;

    case ACE_Asynch_Result_Impl::READ_DGRAM:
      ret_val = this->start_read_dgram (result);
      break;

    case ACE_Asynch_Result_Impl::WRITE_DGRAM:
      ret_val = this->start_write_dgram (result);
      break;

    default:
      ret_val = ACE_POSIX_AIO_Provider::start_not_supported (result);
      break;
    }

  return ret_val;
}


#endif /* ACE_HAS_AIO_CALLS */
