/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

/* -*- C++ -*- */

//=============================================================================
/**
 *
 *  @file    POSIX_AIO_Provider.h
 *
 *  $Id:  $
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef ACE_POSIX_AIO_PROVIDER_H
#define ACE_POSIX_AIO_PROVIDER_H

#include "ace/POSIX_AIO_Strategies.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
#pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#if defined (ACE_HAS_AIO_CALLS) || defined(ACE_HAS_AIO_EMULATION)


/**
 * @class ACE_POSIX_AIO_Provider
 *
 */
class ACE_Export ACE_POSIX_AIO_Provider
{
  /// This interface ( abstract class) defines
  /// what we need from OS to perfom necessary
  /// IO operations

public:

  friend class ACE_POSIX_AIO_Processor;

  enum Cancel_Retcode
    {
      CR_CANCELED     = 0,  // canceled, check result for ready
      CR_ALL_DONE     = 1,  // already finished or canceled
      CR_NOT_CANCELED = 2,  // unable to cancel
      CR_CANCEL_ERROR = -1  // errors
    };

  ACE_POSIX_AIO_Provider (const ACE_POSIX_AIO_Config & config);
  
  virtual ~ACE_POSIX_AIO_Provider(void);

  const ACE_POSIX_AIO_Config & config(void) const;

  /// This function is a no-op function for POSIX/SUN AIO Providers.
  /// But should be implemented for ACE_POSIX_AIO_Reactor_Provider.
  virtual int register_handle (ACE_HANDLE handle,
                               const void *completion_key)=0;

  /// This function is a no-op function for POSIX/SUN AIO Providers.
  /// But should be implemented for ACE_POSIX_AIO_Reactor_Provider.
  virtual int unregister_handle (ACE_HANDLE handle)=0;

  virtual int get_supported_operations_mask () const =0;
  virtual int should_start_interrupt_leader () const; 
  virtual int can_start_in_any_thread () const;
  virtual size_t get_max_aio_num () const;


  ACE_POSIX_AIO_Wait_Strategy * wait_strategy () const;
  ACE_POSIX_AIO_Interrupt_Strategy * interrupt_strategy () const;

protected:

  /// called by AIO_Processor to create 
  /// Wait_Strategy ( desirably of ws_type) and then
  /// Interrupt_Strategy ( desirably of is_type) 
  int create_strategies (void);

  /// called by AIO_Processor to destroy strategies
  int destroy_strategies (void);


  virtual ACE_POSIX_AIO_Wait_Strategy * create_wait_strategy (void) = 0;
  virtual void  destroy_wait_strategy (ACE_POSIX_AIO_Wait_Strategy * w_strategy) =0;

  // Start concrete operations
  virtual int  start_read_stream  (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_write_stream (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_read_file    (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_write_file   (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_read_dgram   (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_write_dgram  (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_accept       (ACE_POSIX_Asynch_Result * result) =0;
  virtual int  start_connect      (ACE_POSIX_Asynch_Result * result) =0;
  

  /// Cancel pending AIO request
  /// Returns Cancel_Retcode
  virtual int cancel_aio (ACE_POSIX_Asynch_Result * result)=0;

  /// Start AIO request
  /// Operation code is saved in ACE_POSIX_Asynch_Result
  /// On return see ACE_POSIX_Asynch_Result::state()
  ///   STARTED
  ///   DEFERRED 
  ///   RESULT_READY
  ///   START_ERROR
  /// return code 
  ///   1 - start should interrupt leader
  ///   0 - start should not interrupt leader
  int  start_aio (ACE_POSIX_Asynch_Result * result);

 
  /// Common useful helper for AIO
  /// Define the maximum number of asynchronous I/O requests
  /// for the current OS ;
  static size_t get_os_max_aio_num (size_t max_op);

  /// Common useful helper
  /// translate rc and errno into ACE_POSIX_Asynch_Result::State
  static void check_start_error  (ACE_POSIX_Asynch_Result *result, int rc);

  /// Common useful helper
  /// what to do with unsupported operations
  static int start_not_supported (ACE_POSIX_Asynch_Result *result);

  ///
  const ACE_POSIX_AIO_Config & config_;

  ///  Maximum number of start AIO for this provider
  size_t max_aio_num_;

  ACE_POSIX_AIO_Wait_Strategy * wait_strategy_;
  ACE_POSIX_AIO_Interrupt_Strategy * interrupt_strategy_;
};


#endif /* ACE_HAS_AIO_CALLS */

#endif /* ACE_POSIX_AIO_PROVIDER_H */
