/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

// ============================================================================
/**
 *  @file PSSL_Test.cpp
 *
 *  PSSL_Test.cpp
 *
 *  This program illustrates how the ACE_Proactor can be used to
 *  implement an application that does various asynchronous
 *  operations.
 *
 *  @author Alexander Libman <libman@terabit.com.au>
 */
// ============================================================================

#include "test_config.h"

#if defined (ACE_HAS_THREADS) && ((defined (ACE_WIN32) && !defined (ACE_HAS_WINCE)) || (defined (ACE_HAS_AIO_CALLS)))
  // This only works on Win32 platforms and on Unix platforms
  // supporting POSIX aio calls.

#include "Cfg.h"
#include "ProactorTask.h"

#include "ace/INET_Addr.h"
#include "ace/Asynch_Acceptor.h"
#include "ace/Asynch_Connector.h"

#include "ace/SSL/sslconf.h"
#include "ace/SSL/SSL_Asynch_Stream.h"
#include "openssl/rand.h"


static ACE_TCHAR complete_message[] =
  ACE_TEXT ("GET / HTTP/1.1\r\n")
  ACE_TEXT ("Accept: */*\r\n")
  ACE_TEXT ("Accept-Language: C++\r\n")
  ACE_TEXT ("Accept-Encoding: gzip, deflate\r\n")
  ACE_TEXT ("User-Agent: PSSL_Test /1.0 (non-compatible)\r\n")
  ACE_TEXT ("Connection: Keep-Alive\r\n")
  ACE_TEXT ("\r\n");


static TestCfg cfg;

// *************************************************************
//   Receiver and Acceptor
// *************************************************************
// forward declaration
class Acceptor;

class Receiver : public ACE_Service_Handler
{
  friend class Acceptor;
public:
  Receiver  (Acceptor *acceptor = 0, int index = -1);
  ~Receiver (void);

  size_t get_total_snd (void)const { return this->total_snd_; }
  size_t get_total_rcv (void)const { return this->total_rcv_; }
  long get_total_w   (void)  const { return this->total_w_; }
  long get_total_r   (void)  const { return this->total_r_; }
  int get_ref_cnt_w (void)   const { return this->ref_cnt_w_; }
  int get_ref_cnt_r (void)   const { return this->ref_cnt_r_; }

  // This is called to pass the new connection's addresses.
  virtual void addresses (const ACE_INET_Addr& peer,
                          const ACE_INET_Addr& local);

  /// This is called after the new connection has been accepted.
  virtual void open (ACE_HANDLE handle,
                     ACE_Message_Block &message_block);

protected:
  /**
   * @name AIO callback handling
   *
   * These methods are called by the framework
   */
  /// This is called when asynchronous <read> operation from the
  /// socket completes.
  virtual void handle_read_stream (const ACE_Asynch_Read_Stream::Result &result);

  /// This is called when an asynchronous <write> to the socket
  /// completes.
  virtual void handle_write_stream (const ACE_Asynch_Write_Stream::Result &result);

  /// This is called after all I/O is done and
  /// it is safe to self-destroy 
  virtual void handle_wakeup ();
  
  void cancel ();
  void close  ();

private:
  int initiate_read_stream (void);
  int initiate_write_stream (ACE_Message_Block &mb, size_t nbytes);
  int has_pending(void);

  Acceptor *acceptor_;
  int index_;

  ACE_SSL_Asynch_Stream stream_;
  ACE_SYNCH_MUTEX lock_;

  int    ref_cnt_r_;         // Number of currently outstanding Read requests
  int    ref_cnt_w_;         // Number of currently outstanding Write requests
  int    flg_cancel_;
  int    flg_safe_delete_;
  size_t total_snd_;         // Number of bytes successfully sent
  size_t total_rcv_;         // Number of bytes successfully received
  long total_w_;             // Number of write operations
  long total_r_;             // Number of read operations
};

class Acceptor : public ACE_Asynch_Acceptor<Receiver>
{
  friend class Receiver;
public:
  int get_number_sessions (void) { return this->sessions_; }
  size_t get_total_snd (void) { return this->total_snd_; }
  size_t get_total_rcv (void) { return this->total_rcv_; }
  long get_total_w   (void) { return this->total_w_; }
  long get_total_r   (void) { return this->total_r_; }

  int  get_number_pending_accepts(void) { return this->ref_cnt_a_;}

  Acceptor (ProactorTask &task);
  virtual ~Acceptor (void);

  ProactorTask & task () { return task_;}
  int  start(const ACE_INET_Addr & addr);
  void stop (void);
  void cancel_all (void);

  int  is_safe_to_destroy (void);

  // Virtual from ACE_Asynch_Acceptor
  Receiver *make_handler (void);

  int  validate_connection(const ACE_Asynch_Accept::Result&  result,
                           const ACE_INET_Addr& remote,
                           const ACE_INET_Addr& local );

  int should_reissue_accept (void);


private:
  void on_new_receiver (Receiver &rcvr);
  void on_delete_receiver (Receiver &rcvr);

  ProactorTask & task_;

  ACE_SYNCH_RECURSIVE_MUTEX lock_;
  u_int sessions_;
  Receiver *list_receivers_[MAX_SESSIONS];

  size_t total_snd_;
  size_t total_rcv_;
  long total_w_;
  long total_r_;
  int  flg_cancel_;

  int  ref_cnt_a_;   // Number of currently outstanding Accept requests
};

// *************************************************************
Acceptor::Acceptor (ProactorTask & task)
  : task_     (task),
    sessions_ (0),
    total_snd_(0),
    total_rcv_(0),
    total_w_  (0),
    total_r_  (0),
    flg_cancel_(0),
    ref_cnt_a_ (0)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
     this->list_receivers_[i] = 0;
}

Acceptor::~Acceptor (void)
{
  this->stop ();
}

int
Acceptor::start(const ACE_INET_Addr & addr)
{
  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_,-1);

  int num_initial_accepts = 5;

  int rc = this->open(addr,
                    0,   //size_t bytes_to_read = 0,
                    1,   //int pass_addresses = 0,
                    ACE_DEFAULT_BACKLOG, //int backlog ,
                    1,   //int reuse_addr = 1,
                    task_.proactor(), // ACE_Proactor *proactor = 0,
                    1,   //int validate_new_connection = 0,
                    0,   //int reissue_accept = 1,
                    0);  //int number_of_initial_accepts = -1);


  for (; rc == 0 && this->ref_cnt_a_ < num_initial_accepts; 
         ++this->ref_cnt_a_)
    {
      rc =this->accept(0,0);
    }

  return rc;
}


void
Acceptor::cancel_all (void)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  this->flg_cancel_ = 1;

  this->reissue_accept(0);  // stop future accepts
  this->cancel ();

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      if (this->list_receivers_[i] != 0)
        this->list_receivers_[i]->cancel ();
    }

  return;
}


void
Acceptor::stop (void)
{
  // This method can be called only after proactor event loop is done
  // in all threads.

  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      delete this->list_receivers_[i];
      this->list_receivers_[i] = 0;
    }
}

void
Acceptor::on_new_receiver (Receiver & rcvr)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);
  this->sessions_++;
  this->list_receivers_[rcvr.index_] = &rcvr;
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t) Acceptor: NEW Receiver=%d total=%d.\n"),
              rcvr.index_,
              this->sessions_));
}

void
Acceptor::on_delete_receiver (Receiver & rcvr)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  this->sessions_--;

  this->total_snd_ += rcvr.get_total_snd();
  this->total_rcv_ += rcvr.get_total_rcv();
  this->total_w_   += rcvr.get_total_w();
  this->total_r_   += rcvr.get_total_r();

  if (rcvr.index_ >= 0
      && (u_int) rcvr.index_ < MAX_SESSIONS
      && this->list_receivers_[rcvr.index_] == &rcvr)
    this->list_receivers_[rcvr.index_] = 0;

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t) Acceptor: DEL Receiver=%d S=%d(%d|%d) R=%d(%d|%d) total=%d\n"),
              rcvr.index_,
              rcvr.get_total_snd(),
              rcvr.get_total_w(),
              rcvr.get_ref_cnt_w(),
              rcvr.get_total_rcv(),
              rcvr.get_total_r(),
              rcvr.get_ref_cnt_r(),
              this->sessions_));

  if (this->sessions_ == 0 && this->ref_cnt_a_ == 0)
    this->task().signal_main();
}

int
Acceptor::validate_connection(const ACE_Asynch_Accept::Result&  result,
                              const ACE_INET_Addr& /*remote*/,
                              const ACE_INET_Addr& /*local*/)
{
  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_,-1);

  --this->ref_cnt_a_;

  int rc = 0;

  if (!result.success())
    {
      ACE_DEBUG ((LM_DEBUG,
                  ACE_TEXT ("(%t) Acceptor: Accept failed\n")));
      rc = -1;
    }
  
  if (this->flg_cancel_ != 0 )
    {
      rc = -1;
    }
    else
    {
      if (this->accept(0,0) == 0)
         ++this->ref_cnt_a_ ;
      else
          ACE_DEBUG ((LM_DEBUG,
                      ACE_TEXT ("(%t) Acceptor: Reissue Accept failed\n")));
    }


  if(this->is_safe_to_destroy())
    this->task().signal_main();

  return rc;
}

int
Acceptor::is_safe_to_destroy (void)
{
  return (this->sessions_ == 0 && this->ref_cnt_a_ == 0);
}

int
Acceptor::should_reissue_accept (void)
{ 
  // we provide manual control for reiussue accept
  // to manage number outstanding accepts properly
  // see validate_connection(...) method 
  
  //if (this->flg_cancel_ != 0)
  //  return 0;

  return 0;
}


Receiver *
Acceptor::make_handler (void)
{
  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_, 0);

  if (this->sessions_ >= MAX_SESSIONS || this->flg_cancel_ != 0)
    return 0;

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      if (this->list_receivers_[i] == 0)
        {
          ACE_NEW_RETURN (this->list_receivers_[i],
                          Receiver (this, i),
                          0);
          return this->list_receivers_[i];
        }
    }

  return 0;
}

// ***************************************************
Receiver::Receiver (Acceptor * acceptor, int index)
  : acceptor_ (acceptor),
    index_    (index),
    stream_   (ACE_SSL_Asynch_Stream::ST_SERVER),
    ref_cnt_r_ (0),
    ref_cnt_w_ (0),
    flg_cancel_(0),
    flg_safe_delete_(0),
    total_snd_(0),
    total_rcv_(0),
    total_w_  (0),
    total_r_  (0)
{
  if (this->acceptor_ != 0)
    this->acceptor_->on_new_receiver (*this);
}

Receiver::~Receiver (void)
{
  if (this->acceptor_ != 0)
    this->acceptor_->on_delete_receiver (*this);


  this->close();
  this->index_ = -1;
}

int 
Receiver::has_pending(void)
{
  return (this->ref_cnt_r_ != 0 ||
           this->ref_cnt_w_ != 0 ||
           flg_safe_delete_ == 0);
}

void
Receiver::cancel ()
{
  ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);
  
  this->flg_cancel_ = 1;
  this->stream_.cancel_and_close();
}

void
Receiver::close ()
{
}

void
Receiver::addresses (const ACE_INET_Addr& peer, const ACE_INET_Addr&)
{
  ACE_TCHAR str[256];
  if (0 == peer.addr_to_string (str, sizeof (str) / sizeof (ACE_TCHAR)))
    ACE_DEBUG ((LM_DEBUG,
                ACE_TEXT ("(%t) Receiver %d connection from %s\n"),
                this->index_,
                str));
  else
    ACE_ERROR ((LM_ERROR, ACE_TEXT ("(%t) Receiver %d %p\n"),
                this->index_,
                ACE_TEXT ("addr_to_string")));
  return;
}


void
Receiver::open (ACE_HANDLE handle, ACE_Message_Block &)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);

    if (this->stream_.open (*this,
                             this->handle_,
                             0, // completion key,
                             this->acceptor_->proactor(),
                             1 ) == -1)
      {
        this->stream_.close();                             
        ACE_ERROR ((LM_ERROR,
                  ACE_TEXT ("%p\n"),
                  ACE_TEXT ("Receiver::ACE_SSL_Asynch_Stream::open")));
      }
    else
      this->initiate_read_stream ();

    if (this->has_pending())
      return;

  }
   
  delete this;
}

int
Receiver::initiate_read_stream (void)
{
  ACE_Message_Block *mb = 0;

  if (this->flg_cancel_ != 0 || this->get_ref_cnt_r() != 0)
     return 0;

  ACE_NEW_RETURN (mb,
                  ACE_Message_Block (cfg.r_blksize()+1),
                  -1);

  // Inititiate read
  if (this->stream_.read (*mb, cfg.r_blksize()) == -1)
    {
      mb->release ();
      this->stream_.cancel();   // close() for SSL shutdown
      return -1;
    }

  this->ref_cnt_r_++;
  this->total_r_++;
  return 0;
}

int
Receiver::initiate_write_stream (ACE_Message_Block &mb, size_t nbytes)
{

  if (this->flg_cancel_ != 0 || nbytes == 0)
    {
      mb.release ();
      this->stream_.cancel();    // close() for SSL shutdown
      return -1;
    }

  if (this->stream_.write (mb, nbytes) == -1)
    {
      mb.release ();
      this->stream_.cancel();    // close() for SSL shutdown
      return -1;
    }

  this->ref_cnt_w_++;
  this->total_w_++;
  return 0;
}

void Receiver::handle_wakeup ()
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_ );

    this->flg_safe_delete_ = 1;
      
    if (this->has_pending())
      return;
  }
  delete this;
  return;
}
void
Receiver::handle_read_stream (const ACE_Asynch_Read_Stream::Result &result)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_ );

    this->ref_cnt_r_--;

    ACE_Message_Block & mb = result.message_block ();

    // Reset pointers.
    mb.rd_ptr ()[result.bytes_transferred ()] = '\0';

    if (cfg.loglevel() == 0)
      {
        LogLocker log_lock;

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) **** Receiver %d: handle_read_stream() ****\n"),
                    this->index_));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_to_read"),
                    result.bytes_to_read ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("handle"),
                    result.handle ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_transfered"),
                    result.bytes_transferred ()));
/*
-------------------------------------------
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("act"),
                    result.act ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("success"),
                    result.success ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("completion_key"),
                    result.completion_key ()));

-------------------------------------------
*/

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("error"),
                    result.error ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %s\n"),
                    ACE_TEXT ("message_block"),
                    mb.rd_ptr ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("**** end of message ****************\n")));
      }
    else if (result.error () != 0)
      {
        LogLocker log_lock;

        ACE_Log_Msg::instance ()->errnum (result.error ());
        ACE_Log_Msg::instance ()->log (LM_ERROR,
                                       ACE_TEXT ("(%t) ERROR Receiver=%d %p\n"),
                                       this->index_,
                                       ACE_TEXT ("read"));
      }
    else if ( cfg.loglevel () ==1 )
      {
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) Receiver=%d read=%d bytes\n"),
                    this->index_,
                    result.bytes_transferred ()));
      }


    do_delay (cfg.delay());  // delay milliseconds

    if (result.error () == 0 && result.bytes_transferred () > 0)
      {
        this->total_rcv_ += result.bytes_transferred ();

        this->initiate_write_stream (mb, result.bytes_transferred ());

        if ((u_long)(this->total_rcv_ - this->total_snd_) <= cfg.w_size())
          this->initiate_read_stream ();
      }
    else
      {
        mb.release ();
      }

    if (result.error () != 0)
        this->stream_.cancel ();

    if (this->ref_cnt_r_ != 0 || this->ref_cnt_w_ !=0)
      return;
  }
  delete this;
}

void
Receiver::handle_write_stream (const ACE_Asynch_Write_Stream::Result &result)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);

    this->ref_cnt_w_--;

    ACE_Message_Block & mb = result.message_block ();

    if (cfg.loglevel () == 0)
      {
        LogLocker log_lock;

        //mb.rd_ptr () [0] = '\0';
        mb.rd_ptr (mb.rd_ptr () - result.bytes_transferred ());

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) **** Receiver %d: handle_write_stream() ****\n"),
                    this->index_));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_to_write"),
                    result.bytes_to_write ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("handle"),
                    result.handle ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_transfered"),
                    result.bytes_transferred ()));
/*
-------------------------------------------
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("act"),
                    result.act ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("success"),
                    result.success ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("completion_key"),
                    result.completion_key ()));
-------------------------------------------
*/
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("error"),
                    result.error ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %s\n"),
                    ACE_TEXT ("message_block"),
                    mb.rd_ptr ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("**** end of message ****************\n")));
      }
    else if (result.error () != 0)
      {
        LogLocker log_lock;

        ACE_Log_Msg::instance ()->errnum (result.error ());
        ACE_Log_Msg::instance ()->log (LM_ERROR,
                                       ACE_TEXT ("(%t) ERROR Receiver=%d %p\n"),
                                       this->index_,
                                       ACE_TEXT ("write"));
      }
    else if ( cfg.loglevel () ==1 )
      {
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) Receiver=%d write=%d bytes ok\n"),
                    this->index_,
                    result.bytes_transferred ()));
      }

    mb.release ();

    if (result.error () == 0 && result.bytes_transferred () > 0)
      {
        u_long delta = (u_long)(this->total_rcv_ - this->total_snd_);

        this->total_snd_ += result.bytes_transferred ();

        if ( delta > cfg.w_size()  && 
             delta <= cfg.w_size() +result.bytes_transferred ())
          this->initiate_read_stream ();
      }
    else
      this->stream_.cancel_and_close ();


    if (this->ref_cnt_r_ != 0 || this->ref_cnt_w_ !=0)
      return;
  }
  delete this;
}

// *******************************************
//   Sender
// *******************************************

class Connector;

class Sender : public ACE_Service_Handler
{
  friend class Connector;
public:

  /// This is called after the new connection has been established.
  virtual void open (ACE_HANDLE handle,
                     ACE_Message_Block &message_block);

  Sender  (Connector *connector = 0, int index = -1);
  ~Sender (void);

  size_t get_total_snd (void)const { return this->total_snd_; }
  size_t get_total_rcv (void)const { return this->total_rcv_; }
  long get_total_w   (void)  const { return this->total_w_; }
  long get_total_r   (void)  const { return this->total_r_; }
  int get_ref_cnt_w (void)   const { return this->ref_cnt_w_; }
  int get_ref_cnt_r (void)   const { return this->ref_cnt_r_; }

  // This is called to pass the new connection's addresses.
  virtual void addresses (const ACE_INET_Addr& peer,
                          const ACE_INET_Addr& local);

  virtual void handle_read_stream (const ACE_Asynch_Read_Stream::Result &result);
  // This is called when asynchronous reads from the socket complete

  virtual void handle_write_stream (const ACE_Asynch_Write_Stream::Result &result);
  // This is called when asynchronous writes from the socket complete

  void close (void);
  void cancel (void);

private:
  int initiate_read_stream (void);
  int initiate_write_stream (void);


  int  index_;
  Connector * connector_;

  Asynch_RW_Stream  stream_;

  ACE_SYNCH_MUTEX lock_;

  int  ref_cnt_r_;
  int  ref_cnt_w_;
  size_t total_snd_;
  size_t total_rcv_;
  long total_w_;
  long total_r_;
};

class Connector : public ACE_Asynch_Connector<Sender>
{
  friend class Sender;
public:
  int get_number_sessions (void) { return this->sessions_; }
  size_t get_total_snd (void) { return this->total_snd_; }
  size_t get_total_rcv (void) { return this->total_rcv_; }
  long get_total_w   (void) { return this->total_w_; }
  long get_total_r   (void) { return this->total_r_; }

  int  get_number_pending_connects(void) { return this->ref_cnt_c_;}

  Connector (ProactorTask & task);
  virtual ~Connector (void);

  ProactorTask & task () { return task_;}

  int  start (const ACE_INET_Addr &addr, u_int num);
  void stop (void);
  void cancel_all (void);
  int  is_safe_to_destroy (void);
  

  // Virtual from ACE_Asynch_Connector
  Sender *make_handler (const void * act);
  int  validate_connection(const ACE_Asynch_Connect::Result&  result,
                           const ACE_INET_Addr& remote,
                           const ACE_INET_Addr& local );

private:
  void on_new_sender  (Sender &rcvr);
  void on_delete_sender (Sender &rcvr);

  ACE_SYNCH_RECURSIVE_MUTEX lock_;
  ProactorTask & task_;
  u_int sessions_;
  Sender *list_senders_[MAX_SESSIONS];
  size_t total_snd_;
  size_t total_rcv_;
  long total_w_;
  long total_r_;
  int  flg_cancel_;

  int  ref_cnt_c_;   // Number of currently outstanding Connect requests
};

// *************************************************************

Connector::Connector (ProactorTask & task)
  : task_     (task),
    sessions_ (0),
    total_snd_(0),
    total_rcv_(0),
    total_w_  (0),
    total_r_  (0),
    flg_cancel_(0),
    ref_cnt_c_ (0)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
     this->list_senders_[i] = 0;
}

Connector::~Connector (void)
{
  this->stop ();
}


void
Connector::cancel_all(void)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  this->flg_cancel_ = 1;

  this->cancel ();

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      if (this->list_senders_[i] != 0)
        this->list_senders_[i]->cancel ();
    }
  return;
}

void
Connector::stop (void)
{
  // This method can be called only after proactor event loop is done
  // in all threads.

  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      delete this->list_senders_[i];
      this->list_senders_[i] = 0;
    }
}

void
Connector::on_new_sender (Sender &sndr)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);
  this->sessions_++;
  this->list_senders_[sndr.index_] = &sndr;
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t) Connector: NEW Sender=%d  total=%d\n"),
              sndr.index_,
              this->sessions_));
}

void
Connector::on_delete_sender (Sender &sndr)
{
  ACE_GUARD (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_);

  this->sessions_--;
  this->total_snd_ += sndr.get_total_snd();
  this->total_rcv_ += sndr.get_total_rcv();
  this->total_w_   += sndr.get_total_w();
  this->total_r_   += sndr.get_total_r();

  if (sndr.index_ >= 0
      && (u_int) sndr.index_ < MAX_SESSIONS
      && this->list_senders_[sndr.index_] == &sndr)
    this->list_senders_[sndr.index_] = 0;

  
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t) Connector: DEL Sender=%d S=%d(%d|%d) R=%d(%d|%d) total=%d\n"),
              sndr.index_,
              sndr.get_total_snd(),
              sndr.get_total_w(),
              sndr.get_ref_cnt_w(),
              sndr.get_total_rcv(),
              sndr.get_total_r(),
              sndr.get_ref_cnt_r(),
              this->sessions_));

  if (this->is_safe_to_destroy())
    this->task().signal_main();
}

Sender *
Connector::make_handler (const void * /*act*/)
{
  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_, 0);

  if (this->sessions_ >= MAX_SESSIONS || this->flg_cancel_ != 0)
      return 0;

  for (u_int i = 0; i < MAX_SESSIONS; ++i)
    {
      if (this->list_senders_ [i] == 0)
        {
          ACE_NEW_RETURN (this->list_senders_[i],
                          Sender (this, i),
                          0);
          return this->list_senders_[i];
        }
    }

  return 0;
}


int
Connector::start (const ACE_INET_Addr& addr, u_int num)
{

  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_, 0);

  if (num > MAX_SESSIONS)
    num = MAX_SESSIONS;

  u_int rc = 0;

  if (this->open (1,  // int pass_addresses = 0,
                  task_.proactor(), //ACE_Proactor *proactor = 0,
                  1    // int validate_new_connection = 0 );
                 )!= 0)
  {
     ACE_ERROR ((LM_ERROR,
                 ACE_LIB_TEXT ("(%t) %p\n"),
                 ACE_LIB_TEXT ("Connector::open failed")));
     return rc;
  }

  for (; rc < num;  ++rc, ++this->ref_cnt_c_ )
    {
      //int connect (const ACE_INET_Addr &remote_sap,
      //             const ACE_INET_Addr &local_sap =
      //                          ACE_INET_Addr ((u_short)0),
      //                 int reuse_addr = 1,
      //                 const void *act = 0);

      if (this->connect (addr,
                         ACE_INET_Addr ((u_short)0),
                         1,
                         (const void *) rc) != 0)
        {
          ACE_ERROR ((LM_ERROR,
                      ACE_TEXT ("(%t) %p\n"),
                      ACE_TEXT ("Connector::connect failed")));
          break;
        }
    }
  return (u_int) rc;
}

int
Connector::validate_connection(const ACE_Asynch_Connect::Result&  result,
                               const ACE_INET_Addr& /*remote*/,
                               const ACE_INET_Addr& /*local*/)
{
  ACE_GUARD_RETURN (ACE_SYNCH_RECURSIVE_MUTEX, monitor, this->lock_, 0);

  --this->ref_cnt_c_;

  int rc = 0;
  if (!result.success())
    {
      rc = -1;

      //ACE_DEBUG ((LM_DEBUG,
      //            ACE_TEXT ("(%t) Connector: Connect failed\n")));
    }

  if(this->is_safe_to_destroy())
    this->task().signal_main();

  return rc;
}

int
Connector::is_safe_to_destroy()
{
  return (this->sessions_ == 0 && this->ref_cnt_c_ == 0);
}

// ************************************************

Sender::Sender (Connector * connector, int index)
  : index_     (index),
    connector_ (connector),
    ref_cnt_r_  (0),
    ref_cnt_w_ (0),
    total_snd_ (0),
    total_rcv_ (0),
    total_w_   (0),
    total_r_   (0)
{
  if (this->connector_ != 0)
    this->connector_->on_new_sender (*this);
}

Sender::~Sender (void)
{
  if (this->connector_ != 0)
    this->connector_->on_delete_sender (*this);

  this->close();
  this->index_ = -1;
}

void
Sender::cancel ()
{
  ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);
  this->stream_.cancel_and_close();
}

void
Sender::close ()
{
  ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);
  this->stream_.close();
  return;
}


void
Sender::addresses (const ACE_INET_Addr& /* peer */, const ACE_INET_Addr& local)
{
  ACE_TCHAR str[256];
  if (0 == local.addr_to_string (str, sizeof (str) / sizeof (ACE_TCHAR)))
    ACE_DEBUG ((LM_DEBUG,
                ACE_TEXT ("(%t) Sender %d connected on %s\n"),
                this->index_,
                str));
  else
    ACE_ERROR ((LM_ERROR, ACE_TEXT ("(%t) Receiver %d %p\n"),
                this->index_,
                ACE_TEXT ("addr_to_string")));
  return;
}


void
Sender::open (ACE_HANDLE handle, ACE_Message_Block &)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);

    if (this->stream_.open (*this, 
                        handle,
                        0, // completion key,
                        this->connector_->proactor(),
                        1) == -1)

      ACE_ERROR ((LM_ERROR,
                  ACE_TEXT ("(%t) %p\n"),
                  ACE_TEXT ("Sender::ACE_Asynch_RW_Stream::open")));

    else
      {
        this->initiate_read_stream ();

        int num = 1+(cfg.w_size()/cfg.s_blksize());

        for (; num > 0; --num)
          this->initiate_write_stream ();
      }

    if (this->ref_cnt_r_ != 0 || this->ref_cnt_w_ !=0)
      return;
  }
  delete this;
}

int
Sender::initiate_write_stream (void)
{
  if (cfg.xfer_limit() != 0 && this->total_snd_ >= cfg.xfer_limit()) 
    {
      //ACE_OS::shutdown(this->handle_, ACE_SHUTDOWN_WRITE);

      return 0;
    }

  ACE_Message_Block *mb = 0;

  ACE_NEW_RETURN (mb,
                  ACE_Message_Block (cfg.s_blksize()+1),
                  -1);

  mb->rd_ptr(mb->base());
  mb->wr_ptr(mb->base() + cfg.s_blksize());

  ACE_OS::strncpy(mb->rd_ptr(), complete_message,cfg.s_blksize());

  *mb->wr_ptr() = '\0'; 
    
  if (this->stream_.write (*mb, mb->length ()) == -1)
    {
      mb->release ();
      this->stream_.cancel_and_close();
      return -1;
    }

  this->ref_cnt_w_++;
  this->total_w_++;
  return 0;
}

int
Sender::initiate_read_stream (void)
{
  ACE_Message_Block *mb = 0;

  if (this->get_ref_cnt_r() != 0 )
    return 0;

  ACE_NEW_RETURN (mb,
                  ACE_Message_Block (cfg.r_blksize()+1),
                  -1);

  // Inititiate read
  if (this->stream_.read (*mb, mb->size () - 1) == -1)
    {
      mb->release ();
      this->stream_.cancel_and_close();
      return -1;
    }

  this->ref_cnt_r_++;
  this->total_r_++;
  return 0;
}

void
Sender::handle_write_stream (const ACE_Asynch_Write_Stream::Result &result)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);

    this->ref_cnt_w_--;

    ACE_Message_Block & mb = result.message_block ();

    if (cfg.loglevel () == 0)
      { 
        LogLocker log_lock;

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) **** Sender %d: handle_write_stream() ****\n"),
                    index_));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_to_write"),
                    result.bytes_to_write ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("handle"),
                    result.handle ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_transfered"),
                    result.bytes_transferred ()));

/*
-------------------------------------------
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("act"),
                    result.act ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("success"),
                    result.success ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("completion_key"),
                    result.completion_key ()));
-------------------------------------------
*/

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("error"),
                    result.error ()));

        mb.rd_ptr()[0]  = '\0';

        // move rd_ptr backwards as required for printout
        mb.rd_ptr (- (int)result.bytes_transferred ());

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %s\n"),
                    ACE_TEXT ("message_block"),
                    mb.rd_ptr ()));

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("**** end of message ****************\n")));
      }
    else if (result.error () != 0)
      {
        LogLocker log_lock;

        ACE_Log_Msg::instance ()->errnum (result.error ());
        ACE_Log_Msg::instance ()->log (LM_ERROR,
                                       ACE_TEXT ("(%t) ERROR Sender=%d %p\n"),
                                       this->index_,
                                       ACE_TEXT ("write"));
      }
    else if ( cfg.loglevel () ==1 )
      {
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) Sender=%d write=%d bytes ok\n"),
                    this->index_,
                    result.bytes_transferred ()));
      }

    mb.release ();

    if (result.error () == 0 && result.bytes_transferred () > 0)
      {
        this->total_snd_ += result.bytes_transferred ();

        u_long delta = this->total_snd_- this->total_rcv_;

        //for (;delta <= w_size; delta += s_blksize)
        if (delta <= cfg.w_size())
          this->initiate_write_stream ();

        this->initiate_read_stream();
      }
    else
      this->stream_.cancel_and_close ();


    if (this->ref_cnt_r_ != 0 || this->ref_cnt_w_ !=0)
      return;
  }
  delete this;
}

void
Sender::handle_read_stream (const ACE_Asynch_Read_Stream::Result &result)
{
  {
    ACE_GUARD (ACE_SYNCH_MUTEX, monitor, this->lock_);

    this->ref_cnt_r_--;

    ACE_Message_Block & mb = result.message_block ();

    if (cfg.loglevel () == 0)
      {
        LogLocker log_lock;

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) **** Sender %d: handle_read_stream() ****\n"),
                    index_));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_to_read"),
                    result.bytes_to_read ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("handle"),
                    result.handle ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("bytes_transfered"),
                    result.bytes_transferred ()));
/*
-------------------------------------------
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("act"),
                    result.act ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("success"),
                    result.success ()));
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %@\n"),
                    ACE_TEXT ("completion_key"),
                    result.completion_key ()));
-------------------------------------------
*/
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %d\n"),
                    ACE_TEXT ("error"),
                    result.error ()));

        // write 0 at string end for proper printout
        mb.rd_ptr()[result.bytes_transferred ()]  = '\0'; 

        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("%s = %s\n"),
                    ACE_TEXT ("message_block"),
                    mb.rd_ptr ()));


        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("**** end of message ****************\n")));
      }
    else if (result.error () != 0)
      {
        LogLocker log_lock;
        ACE_Log_Msg::instance ()->errnum (result.error ());
        ACE_Log_Msg::instance ()->log (LM_ERROR,
                                       ACE_TEXT ("(%t) ERROR Sender=%d %p\n"),
                                       this->index_,
                                       ACE_TEXT ("read"));
      }
    else if ( cfg.loglevel () ==1 )
      {
        ACE_DEBUG ((LM_DEBUG,
                    ACE_TEXT ("(%t) Sender=%d read=%d bytes ok\n"),
                    this->index_,
                    result.bytes_transferred ()));
      }

    mb.release ();

    if (result.error () == 0 && result.bytes_transferred () > 0)
      {
        this->initiate_read_stream ();

        u_long delta = (this->total_snd_- this->total_rcv_);

        this->total_rcv_ += result.bytes_transferred ();

        if ( delta > cfg.w_size() &&
             delta <= cfg.w_size() +result.bytes_transferred () )
          this->initiate_write_stream ();
      }
    else
      this->stream_.cancel_and_close ();

    if (this->ref_cnt_r_ != 0 || this->ref_cnt_w_ !=0)
      return;
  }
  delete this;
}

// *************************************************************
//   Configuration helpers
// *************************************************************

int
ACE_TMAIN (int argc, ACE_TCHAR *argv[])
{
  ACE_START_TEST (ACE_TEXT ("PSSL_Test"));


  if (cfg.parse_args (argc, argv) < 0)
    return -1;

  disable_signal (ACE_SIGRTMIN, ACE_SIGRTMAX);
  disable_signal (SIGPIPE, SIGPIPE);
  disable_signal (SIGIO, SIGIO);

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("FD_SETSIZE=%d ACE_FDSETSIZE=%d\n"),
              FD_SETSIZE,
              ACE_FD_SETSIZE));

  ACE_DEBUG ((LM_DEBUG,
             "s_blksize=%u r_blksize=%u win_size=%u\n",
              cfg.s_blksize(),
              cfg.r_blksize(),
              cfg.w_size()));


  {
  int rc = 0;

  ProactorTask    task1(cfg);
  Acceptor  acceptor(task1);
  Connector connector(task1);

  ACE_Time_Value time_begin = ACE_OS::gettimeofday ();
  ACE_Time_Value time_end = ACE_OS::gettimeofday ();
  ACE_Time_Value time_run;

  if (task1.start () == 0)
    {
      if (cfg.both() != 0 || cfg.sessions() == 0) // Acceptor
        {
          // Simplify, initial read with zero size
          if (acceptor.start (ACE_INET_Addr (cfg.listen_port())) == 0)
            rc = 1;
        }

      if (cfg.both() != 0 || cfg.sessions() > 0) // Connector
        {
          ACE_INET_Addr addr;

          addr.set (cfg.connect_port(), cfg.connect_host());

          rc += connector.start (addr, cfg.sessions());
        }

      time_begin = ACE_OS::gettimeofday ();

      if (rc > 0)
        {
          task1.enable_event_loop();
          
          ACE_Time_Value  sleep_time (cfg.seconds());
           
          while ( sleep_time != ACE_Time_Value::zero)
            {
              ACE_Countdown_Time countdown ( & sleep_time );

              ACE_OS::sleep (sleep_time );
            }


          if (cfg.seconds () == 0)
            {
              // comment this later
              char c;
              cout << "Press any key to stop=>" << flush;
              cin.clear ();
              cin >> c;
            }
        }
    }



  u_int  num_ssn_s = connector.get_number_sessions ();
  u_int  num_ssn_r = acceptor.get_number_sessions ();

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t)*** Cancel Connector: PendingOP=%d Senders=%d\n"),
              connector.get_number_pending_connects(),
              num_ssn_s
            ));
 
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t)*** Cancel Acceptor: PendingOP=%d Receivers=%d\n"),
              acceptor.get_number_pending_accepts(),
              num_ssn_r
            ));

  // Cancel all pending AIO on Connector and Senders
  // Cancel all pending AIO on Acceptor And Receivers
  connector.cancel_all ();
  acceptor.cancel_all ();
   

  while ( !connector.is_safe_to_destroy() ||
          !acceptor.is_safe_to_destroy())
    {
        task1.wait_signal ();
    }

  time_end = ACE_OS::gettimeofday ();
  time_run = time_end - time_begin;

  u_long snd_bytes_s = connector.get_total_snd();
  u_long rcv_bytes_s = connector.get_total_rcv();
  u_long snd_op_s = connector.get_total_w();
  u_long rcv_op_s = connector.get_total_r();

  u_long snd_bytes_r = acceptor.get_total_snd();
  u_long rcv_bytes_r = acceptor.get_total_rcv();
  u_long snd_op_r = acceptor.get_total_w();
  u_long rcv_op_r = acceptor.get_total_r();

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t) Stop Thread Pool Task\n")));

  task1.stop ();
 

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t)*** Stop Connector: PendingOP=%d Senders=%d\n"),
              connector.get_number_pending_connects(),
              connector.get_number_sessions ()
            ));
 
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("(%t)*** Stop Acceptor: PendingOP=%d Receivers=%d\n"),
              acceptor.get_number_pending_accepts(),
              acceptor.get_number_sessions ()
            ));

  // Now should be safe to destroy
  connector.stop ();
  acceptor.stop ();

  //Print statistic
  ACE_TCHAR bufs [256];
  ACE_TCHAR bufr [256];

  ACE_OS::sprintf (bufs, ACE_TEXT ("%lu(%lu)"),
                   snd_bytes_s,
                   snd_op_s);

  ACE_OS::sprintf (bufr, ACE_TEXT ("%lu(%lu)"),
                   rcv_bytes_s,
                   rcv_op_s);

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Connector/Senders  total bytes: snd=%s rcv=%s\n"),
              bufs,
              bufr));

  ACE_OS::sprintf (bufs, ACE_TEXT ("%lu(%lu)"),
                   snd_bytes_r,
                   snd_op_r);

  ACE_OS::sprintf (bufr, ACE_TEXT ("%lu(%lu)"),
                   rcv_bytes_r,
                   rcv_op_r);
  
  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Acceptor/Receivers total bytes: snd=%s rcv=%s\n"),
              bufs,
              bufr));

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Threads = %d BlkSize=%d\n"),
              cfg.threads(),
              cfg.blksize()));

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Sessions = %d (%d) \n"),
              rc-1,
              cfg.sessions()));


  float  fsec = (float)(time_run.msec()/1000);
  double fspeed = ((float) snd_bytes_s
                  +(float) snd_bytes_r
                  +(float) rcv_bytes_s
                  +(float) rcv_bytes_r
                  );

  if ( fsec > 0.1 )
    fspeed /= fsec; 

  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Effective run time = %f sec\n"),
              fsec));


  ACE_DEBUG ((LM_DEBUG,
              ACE_TEXT ("Effective bytes/sec = %f\n"),
              fspeed));

 }

  ACE_END_TEST;

  return 0;
}

#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)

template class ACE_Asynch_Acceptor<Receiver>;
template class ACE_Asynch_Connector<Sender>;

#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)

#pragma instantiate ACE_Asynch_Acceptor<Receiver>
#pragma instantiate ACE_Asynch_Connector<Sender>

#endif /* ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION */

#else

int
main (int, ACE_TCHAR *[])
{
  ACE_START_TEST (ACE_TEXT ("PSSL_Test"));

  ACE_DEBUG ((LM_INFO,
              ACE_TEXT ("Threads or Asynchronous IO is unsupported.\n")
              ACE_TEXT ("PSSL_Test will not be run.")));

  ACE_END_TEST;

  return 0;
}

#endif  /* ACE_WIN32 && !ACE_HAS_WINCE || ACE_HAS_AIO_CALLS */
