/*********************************************************************
** Copyright (C) 2003 Terabit Pty Ltd.  All rights reserved.
**
** This file is part of the POSIX-Proactor module.
**
** This file may be distributed and used only under the terms of the  
** Terabit Public License as defined by Terabit Pty Ltd of Australia   
** and appearing in the file tlicense.txt included in the packaging of
** this module and available at http://www.terabit.com.au/license.php.
**
** Contact support@terabit.com.au if any conditions of this licensing
** are not clear to you.
**********************************************************************/

// ============================================================================
//
// = LIBRARY
//    
//
// = FILENAME
//    PSession.h
//
// = DESCRIPTION
//
// = AUTHOR
//    Alexander Libman <libman@terabit.com.au>
//
// ============================================================================
#ifndef PSESSION_H
#define PSESSION_H

#include "ProactorTask.h"

#include "ace/Asynch_Acceptor.h"
#include "ace/Asynch_Connector.h"
#include "ace/SString.h"

class PSession;
class PSessionManager;
class PSessionFactory;

//---------------------------------------------------------------------------
// 
//---------------------------------------------------------------------------
class PSessionManager : public ACE_Handler
{
  friend class PSession;

public:
  int    get_number_sessions (void) const { return this->num_sessions_; }
  int    get_peak_sessions (void) const { return this->peak_sessions_; }

  // Sessions statistic
  size_t get_total_snd (void) const { return this->total_snd_; }
  size_t get_total_rcv (void) const { return this->total_rcv_; }
  long   get_total_w   (void) const { return this->total_w_; }
  long   get_total_r   (void) const { return this->total_r_; }

  const ACE_TCHAR * get_name (void) const { return this->name_.c_str();}

  ACE_SYNCH_RECURSIVE_MUTEX  & mutex(void) { return this->lock_;}

  ProactorTask & task (void) { return task_;}

  int    should_finish (void) const { return this->flg_cancel_;}
  int    is_safe_to_delete (void);
 
  // cancel method propagates cancel command to all 
  // active sessions and prevents creation of new sessions
  void   cancel (void);


  PSession * create_session (void);
  void       destroy_session (PSession * session);
  void       destroy_session (int index);

  PSessionManager (ProactorTask &    task, 
                   PSessionFactory & factory,
                   const ACE_TCHAR * name);

  virtual ~PSessionManager  (void);

  void set_timeout (const ACE_Time_Value & timeout);

  const ACE_Time_Value& get_timeout (void) const
    {
      return this->timeout_;
    }

  // virtual from ACE_Handler
  void handle_time_out (const ACE_Time_Value & /* tv */,
                        const void *   /* pArg */);

  void print_statistic(void);

protected:
  // evictor pattern
  void check  (void);

  void cancel_timer (void);
  void start_timer (void);

  ACE_SYNCH_RECURSIVE_MUTEX lock_;
  ProactorTask &            task_;
  PSessionFactory &         factory_;
  ACE_TString               name_;


  size_t total_snd_;  // Total bytes successfully sent
  size_t total_rcv_;  // Total bytes successfully received
  long   total_w_;    // Number of performed write operations
  long   total_r_;    // Number of performed read operations

  int    flg_cancel_; // Request to cancel

private:
  u_int                     num_sessions_;
  u_int                     peak_sessions_;
  PSession *                list_sessions_[MAX_SESSIONS];
  ACE_Time_Value            timeout_;
  long                      timer_id_;
};

//---------------------------------------------------------------------------
// 
//---------------------------------------------------------------------------

class PSession : public ACE_Service_Handler
{
  friend class PSessionManager;

public:
  PSession  (void);
  ~PSession (void);

  size_t get_total_snd (void) const { return this->total_snd_; }
  size_t get_total_rcv (void) const { return this->total_rcv_; }
  long   get_total_w   (void) const { return this->total_w_; }
  long   get_total_r   (void) const { return this->total_r_; }
  int    get_ref_cnt_w (void) const { return this->ref_cnt_w_; }
  int    get_ref_cnt_r (void) const { return this->ref_cnt_r_; }


  ACE_SYNCH_MUTEX & mutex (void) { return this->lock_;}

  int    index (void) const { return this->index_; }

  PSessionManager * manager (void) const { return this->manager_; }

  ProactorTask &  task (void) const { return this->manager()->task();}

  TestCfg & config (void) const { return this->task().config(); }


  int    has_pending_io (void) const
    { 
      return (get_ref_cnt_r() != 0 || get_ref_cnt_w() != 0);
    }

  
  // this should be pure virtual
  // get session name
  virtual const ACE_TCHAR * get_name (void) const
    { throw 0;}

  // this method is called when we have to force
  // cancellation of all outstanding requests 
  virtual void  cancel () 
    { throw 0;}

  // this method is periodically called by PSessionManager
  // when timeout is expired
  virtual void  check ();
  
  virtual int  is_safe_to_delete (void) const;


  // This is called to pass the new connection's addresses.
  virtual void addresses (const ACE_INET_Addr & peer,
                          const ACE_INET_Addr & local);

  void update_last_time (void)
    {
      this->last_op_time_  = ACE_OS::gettimeofday ();
    }

  void set_timeout (const ACE_Time_Value & timeout)
    {
      this->timeout_ = timeout;
    }

  const ACE_Time_Value& get_timeout (void) const
    {
      return this->timeout_;
    }

protected:

  void trace_read_completion  (const ACE_Asynch_Read_Stream::Result &result);
  void trace_write_completion (const ACE_Asynch_Write_Stream::Result &result);

  void set_manager(PSessionManager * manager, int index);

  ACE_SYNCH_MUTEX lock_;

  PSessionManager * manager_;
  int    index_;


  int    ref_cnt_r_;         // Number of currently outstanding Read requests
  int    ref_cnt_w_;         // Number of currently outstanding Write requests
  size_t total_snd_;         // Total bytes successfully sent
  size_t total_rcv_;         // Total bytes successfully received
  long   total_w_;           // Number of performed write operations
  long   total_r_;           // Number of performed read operations
  ACE_Time_Value    last_op_time_;  // time of last activity
  ACE_Time_Value    timeout_;       // timeout value (0 means infinite)
};

//---------------------------------------------------------------------------
// 
//---------------------------------------------------------------------------

class PSessionFactory 
{
public:
  virtual PSession*  create_session(PSessionManager & manager)=0;
};

//---------------------------------------------------------------------------
// 
//---------------------------------------------------------------------------
class Acceptor :  public ACE_Asynch_Acceptor<PSession>
{
public:
  Acceptor (PSessionManager & manager);
  virtual ~Acceptor (void);


  //  pending operation counter
  int    get_ref_cnt  (void) const   { return this->ref_cnt_; }
  int    should_finish (void) const  { return flg_cancel_;}

  int    is_safe_to_delete (void);


  PSessionManager & manager (void)  { return this->manager_; }
  const PSessionManager & manager (void) const  { return this->manager_; }

  ProactorTask & task (void)        { return manager().task();}

  ACE_SYNCH_RECURSIVE_MUTEX  & mutex (void) { return manager().mutex();}

  
  int    start(const ACE_INET_Addr & addr);


    
  // Virtual from ACE_Asynch_Acceptor
  virtual int cancel (void);

  virtual PSession * make_handler (void);

  virtual int  validate_connection(const ACE_Asynch_Accept::Result&  result,
                           const ACE_INET_Addr& remote,
                           const ACE_INET_Addr& local );

  virtual int should_reissue_accept (void);


protected:
  PSessionManager & manager_;
  int               ref_cnt_;    // Number of currently outstanding accepts
  int               flg_cancel_;
};

//---------------------------------------------------------------------------
// 
//---------------------------------------------------------------------------
class Connector : public ACE_Asynch_Connector<PSession>
{
public:
  Connector (PSessionManager & manager);
  virtual ~Connector (void);

  //  pending operation counter
  int    get_ref_cnt  (void) const  { return this->ref_cnt_; }
  
  int    should_finish(void) const  { return flg_cancel_;}

  int    is_safe_to_delete (void);


  PSessionManager & manager (void)  { return this->manager_; }
  const PSessionManager & manager (void) const  { return this->manager_; }

  ProactorTask & task (void)        { return manager().task();}

  ACE_SYNCH_RECURSIVE_MUTEX  & mutex (void) { return manager().mutex();}


  int    start (const ACE_INET_Addr &addr, u_int num);

  
  // Virtual from ACE_Asynch_Connector
  virtual int cancel (void);

  virtual PSession * make_handler (void);

  virtual int validate_connection(const ACE_Asynch_Connect::Result&  result,
                           const ACE_INET_Addr& remote,
                           const ACE_INET_Addr& local );

protected:
  PSessionManager & manager_;
  int               ref_cnt_;    // Number of currently outstanding connects
  int               flg_cancel_;
};

#endif //PSESSION_H
