// -*- C++ -*-

//=============================================================================
/**
 *  @file    SSL_Acceptor.h
 *
 *  SSL_Acceptor.h,
 *
 *
 *  @author Douglas C. Schmidt <schmidt@cs.wustl.edu>
 *
 *   author Alexander Libman <libman@terabit.com.au>
 */
//=============================================================================

#ifndef ACE_SSL_ACCEPTOR_H
#define ACE_SSL_ACCEPTOR_H

#include "ace/pre.h"

#include "ace/Service_Object.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
# pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

#include "ace/Acceptor.h"


/**
 * @class ACE_SSL_Acceptor
 *
 * @brief Abstract factory for creating a service handler
 * (SVC_HANDLER), accepting into the SVC_HANDLER, and
 * activating the SVC_HANDLER.
 *
 * Implements the basic strategy for passively establishing
 * connections with clients.  An ACE_SSL_Acceptor is parameterized
 * by concrete types that conform to the interfaces of
 * PEER_ACCEPTOR and SVC_HANDLER.  The PEER_ACCEPTOR is
 * instantiated with a transport mechanism that passively
 * establishes connections.  The SVC_HANDLER is instantiated
 * with a concrete type that performs the application-specific
 * service.  An ACE_SSL_Acceptor inherits from ACE_Service_Object,
 * which in turn inherits from ACE_Event_Handler.  This enables
 * the ACE_Reactor to dispatch the ACE_SSL_Acceptor's handle_input
 * method when connection events occur.  The handle_input method
 * performs the ACE_SSL_Acceptor's default creation, connection
 * establishment, and service activation strategies.  These
 * strategies can be overridden by subclasses individually or as
 * a group.
 */


template <class SVC_HANDLER, ACE_PEER_ACCEPTOR_1>
class ACE_SSL_Acceptor : 
    public ACE_Acceptor <SVC_HANDLER, ACE_PEER_ACCEPTOR_2>
{

public:



  /// "Do-nothing" constructor.
  ACE_SSL_Acceptor (ACE_Reactor * = 0,
                int use_select = 1);

  /**
   * Initialize and register <this> with the Reactor and listen for
   * connection requests at the designated <local_addr>.  <flags>
   * indicates how <SVC_HANDLER>'s should be initialized prior to
   * being activated.  Right now, the only flag that is processed is
   * <ACE_NONBLOCK>, which enabled non-blocking I/O on the
   * <SVC_HANDLER> when it is opened.  If <use_select> is non-zero
   * then <select> is used to determine when to break out of the
   * <accept> loop.  <reuse_addr> is passed down to the
   * <PEER_ACCEPTOR>.  If it is non-zero this will allow the OS to
   * reuse this listen port.
   */
  ACE_SSL_Acceptor (const ACE_PEER_ACCEPTOR_ADDR &local_addr,
                ACE_Reactor * = ACE_Reactor::instance (),
                int flags = 0,
                int use_select = 1,
                int reuse_addr = 1);

  
  /// Close down the Acceptor's resources.
  virtual ~ACE_SSL_Acceptor (void);

  int deferred_activate (SVC_HANDLER * svc_handler);
  

  /// Declare the dynamic allocation hooks.
  ACE_ALLOC_HOOK_DECLARE;

protected:

  int make_and_activate_ssl_handler (SVC_HANDLER * svc_handler);

  /// Accepts all pending connections from clients, and creates and
  /// activates SVC_HANDLERs.
  virtual int handle_input (ACE_HANDLE);


  /// Default version returns address info in <buf>.
  virtual int info (ACE_TCHAR **buf, size_t) const;

};

template <class SVC_HANDLER, ACE_PEER_ACCEPTOR_1>
class SSL_SVC_HANDLER : public ACE_Event_Handler
{
  friend class ACE_SSL_Acceptor<SVC_HANDLER, ACE_PEER_ACCEPTOR_2>;

 public:

  SSL_SVC_HANDLER  (ACE_SSL_Acceptor <SVC_HANDLER, ACE_PEER_ACCEPTOR_2> * acceptor,
                    SVC_HANDLER * svc_handler);

  ~SSL_SVC_HANDLER ();

  int open();

  virtual ACE_HANDLE get_handle (void) const
    {
      return this->svc_handler_->get_handle();
    }


  virtual void set_handle (ACE_HANDLE h)
    {
      this->svc_handler_->set_handle(h);
    }

  virtual int handle_input (ACE_HANDLE fd = ACE_INVALID_HANDLE);

  virtual int handle_output (ACE_HANDLE fd = ACE_INVALID_HANDLE);

  virtual int handle_close (ACE_HANDLE handle,
                          ACE_Reactor_Mask close_mask);

private:

  int do_ssl_handshake ();

  int set_mask (ACE_Reactor_Mask new_mask);

  // acceptor
  ACE_SSL_Acceptor <SVC_HANDLER, ACE_PEER_ACCEPTOR_2> * acceptor_;

  // user handler
  SVC_HANDLER *    svc_handler_;  

  // flag SSL handshake successfull
  int              success_;  

  // current active mask
  ACE_Reactor_Mask mask_;

  // flag ACE_NONBLOCK passed from the Acceptor
  //int flags_;
};

#if defined (ACE_TEMPLATES_REQUIRE_SOURCE)
#include "SSL_Acceptor.cpp"
#endif /* ACE_TEMPLATES_REQUIRE_SOURCE */

#if defined (ACE_TEMPLATES_REQUIRE_PRAGMA)
#pragma implementation ("SSL_Acceptor.cpp")
#endif /* ACE_TEMPLATES_REQUIRE_PRAGMA */

#include "ace/post.h"

#endif /* ACE_SSL_ACCEPTOR_H */
